# -*- coding: utf-8 -*-
import os
import sys

import unittest

import numpy as np

# temporary solution for relative imports in case pyod is not installed
# if suod
# is installed, no need to use the following line
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from sel_suod.models.base import sel_SUOD

from pyod.utils.data import generate_data
from pyod.models.lof import LOF
from pyod.models.pca import PCA
from pyod.models.hbos import HBOS
from pyod.models.lscp import LSCP
from sel_suod.models.cost_predictor import build_cost_predictor
import joblib


class TestBASE(unittest.TestCase):
	def setUp(self):
		self.n_train = 1000
		self.n_test = 500
		self.contamination = 0.1
		self.roc_floor = 0.6
		self.random_state = 42
		self.X_train, self.X_test, self.y_train, self.y_test = generate_data(
			n_train=self.n_train, n_test=self.n_test, behaviour='new',
			contamination=self.contamination, random_state=self.random_state)

		self.base_estimators = [
			LOF(n_neighbors=5, contamination=self.contamination)
		]
		self.subspaces = np.array([[1,1],[1,1],[1,1],[1,1]])
		self.model = sel_SUOD(base_estimators=self.base_estimators, subspaces=self.subspaces, 
						  n_jobs=2,
						  rp_flag_global=True, bps_flag=True,
						  contamination=self.contamination,
						  approx_flag_global=False,
						  verbose=True)

	def test_initialization(self):
		self.model.get_params()
		self.model.set_params(**{'n_jobs': 4})

	def test_fit(self):
		"""
		Test base class initialization

		:return:
		"""
		self.model.fit(self.X_train)

	def test_approximate(self):
		self.model.fit(self.X_train)
		self.model.approximate(self.X_train)

	def test_predict(self):
		self.model.fit(self.X_train)
		self.model.approximate(self.X_train)
		self.model.predict(self.X_test)

	def test_decision_function(self):
		self.model.fit(self.X_train)
		self.model.approximate(self.X_train)
		self.model.decision_function(self.X_test)

	def test_predict_proba(self):
		self.model.fit(self.X_train)
		self.model.approximate(self.X_train)
		self.model.predict_proba(self.X_test)



# Author: Yue Zhao <zhaoy@cmu.edu>
# License: MIT
from sklearn.base import clone

import numpy as np
from scipy.stats import rankdata
from joblib import effective_n_jobs
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted

from pyod.utils.utility import score_to_label

from .jl_projection import jl_fit_transform
from .sub_projection import sub_transform
from ..utils.utility import raw_score_to_proba


def indices_to_one_hot(data, nb_classes):
    """Convert an iterable of indices to one-hot encoded labels."""
    targets = np.array(data).reshape(-1)
    return np.eye(nb_classes)[targets]


def balanced_scheduling(time_cost_pred, n_estimators, n_jobs, verbose=False):
    """Conduct balanced scheduling based on the sum of rank, for both train
    and prediction. The algorithm will enforce the equal sum of ranks among 
    workers.

    Parameters
    ----------
    time_cost_pred : list
        The list of time cost by the cost predictor. The length is equal to
        the number of base detectors.

    n_estimators : int
        The number of base estimators.

    n_jobs : optional (default=1)
        The number of jobs to run in parallel for both `fit` and
        `predict`. If -1, then the number of jobs is set to the
        number of cores.

    verbose : bool, optional (default=False)
        Controls the verbosity of the building process.

    Returns
    -------
    n_estimators_list : list
        The number of estimators for each worker

    starts : list
        The actual index of base detectors to be scheduled. For instance,
        starts[k, k+1] base detectors will be assigned to worker k.

    n_jobs :
        The actual usable number of jobs to run in parallel.
    """
    # get the number of usable workers
    n_jobs = min(effective_n_jobs(n_jobs), n_estimators)

    # conduct Balanced Task Scheduling
    n_estimators_list = []  # track the number of estimators for each worker
    ranks = rankdata(time_cost_pred)
    ##########################################
    # #todo: the fastest is at most 2 times costly than the slowest
    # enable a parameter for the rank strength
    ranks = 1 + ranks / n_estimators
    ##########################################
    rank_sum = np.sum(ranks)
    chunk_sum = rank_sum / n_jobs

    starts_orig = [0]
    index_track = 0
    sum_check = []

    for i in range(len(ranks) + 1):
        if np.sum(ranks[starts_orig[index_track]:i]) >= chunk_sum:
            starts_orig.append(i)
            index_track += 1
    starts_orig.append(len(ranks))

    starts = starts_orig

    # # offset for the last worker's load
    # starts = [0]
    # for k in range(1, n_jobs+1):
    #     starts.append(starts_orig[k]-np.random.randint(low=1, high=k+1))

    # print(starts)
    # starts[-1] = n_estimators
    # print(starts)

    for j in range(n_jobs):
        sum_check.append(np.sum(ranks[starts[j]:starts[j + 1]]))
        if verbose:
            print('Worker', j + 1, 'sum of ranks:', sum_check[j])
        n_estimators_list.append(starts[j + 1] - starts[j])

    print()

    # Confirm the length of the estimators is consistent
    assert (np.sum(n_estimators_list) == n_estimators)
    assert (np.abs(rank_sum - np.sum(sum_check)) < 0.1)

    xdiff = [starts[n] - starts[n - 1] for n in range(1, len(starts))]

    if verbose:
        print("Split among workers BPS:", starts, xdiff)
    return n_estimators_list, starts, n_jobs


def _partition_estimators(n_estimators, n_jobs, verbose=False):
    """Private function used to partition estimators between jobs.
    """
    # Compute the number of jobs
    n_jobs = min(effective_n_jobs(n_jobs), n_estimators)

    # Partition estimators between jobs
    n_estimators_per_job = np.full(n_jobs, n_estimators // n_jobs,
                                   dtype=int)
    n_estimators_per_job[:n_estimators % n_jobs] += 1
    starts = np.cumsum(n_estimators_per_job)

    xdiff = [starts[n] - starts[n - 1] for n in range(1, len(starts))]

    if verbose:
        print("Split among workers default:", starts, xdiff)
    return n_estimators_per_job.tolist(), [0] + starts.tolist(), n_jobs


def cost_forecast_meta(clf, X, base_estimator_names):
    """Forecast model cost by pretrained cost estimator.

    Parameters
    ----------
    clf : object, sklearn regressor
        Random forest regressor trained to forecast model cost

    X : numpy array of shape (n_samples, n_features)
        The input samples.

    base_estimator_names : list of str
        The list of outlier detection model names in the string format

    Returns
    -------
    time_cost_pred : numpy array of outlier detection model cost in seconds.
    """
    # convert base estimators to the digestible form
    clf_idx = np.asarray(
        list(map(clf_idx_mapping.get, base_estimator_names)))

    X_detector_code = indices_to_one_hot(clf_idx - 1, 11)
    X_shape_code_s = np.array([X.shape[0], X.shape[1]]).reshape(1, 2)
    X_shape_code = np.repeat(X_shape_code_s, len(base_estimator_names), axis=0)
    X_code = np.concatenate((X_shape_code, X_detector_code), axis=1)
    time_cost_pred = clf.predict(X_code)

    return time_cost_pred


def _parallel_fit(n_estimators, clfs, X, total_n_estimators, subspaces,
                  rp_flags, verbose):
    '''
    Main way of fitting an method in all subspaces at the same time
    '''
    X = check_array(X)
    # Build estimators
    estimators = []
    for i, subspace in enumerate(subspaces):
        estimator = clone(clfs[i])
        if verbose > 1:
            print("Building estimator %d of %d for this parallel run "
                  "(total %d)..." % (i + 1, n_estimators, total_n_estimators))

        if rp_flags[i] == 1:
            X_scaled = sub_transform(X, subspace) #Here the projection into the subspace happen.
            

            estimator.fit(X_scaled)
            estimators.append(estimator)
        else:
            estimator.fit(X)
            estimators.append(estimator)
    return estimators


def _parallel_predict(n_estimators, clfs, approximators, X, subspaces, total_n_estimators,
                      approx_flags, contamination, verbose):
    X = check_array(X)

    pred = []
    for i, subspace in enumerate(subspaces):
        estimator = clfs[i]
        if verbose > 1:
            print("predicting with estimator %d of %d for this parallel run "
                  "(total %d)..." % (i + 1, n_estimators, total_n_estimators))

        # project matrix
        X_scaled = sub_transform(X, subspace)

        # turn approximator scores to labels by outlier
        if approx_flags[i] == 1:
            predicted_labels = score_to_label(
                approximators[i].predict(X_scaled),
                outliers_fraction=contamination)

        else:
            predicted_labels = estimator.predict(X_scaled)

        pred.append(predicted_labels)

    return pred


def _parallel_decision_function(n_estimators, clfs, approximators, X,
                                total_n_estimators, subspaces,
                                approx_flags, verbose):
    X = check_array(X)

    pred = []
    for i, subspace in enumerate(subspaces):
        estimator = clfs[i]
        if verbose > 1:
            print("predicting with estimator %d of %d for this parallel run "
                  "(total %d)..." % (i + 1, n_estimators, total_n_estimators))

        # project matrix
        X_scaled = sub_transform(X, subspace)

        # turn approximator scores to labels by outlier
        if approx_flags[i] == 1:
            predicted_scores = approximators[i].predict(X_scaled)
        else:
            predicted_scores = estimator.decision_function(X_scaled)

        pred.append(predicted_scores)

    return pred


def _parallel_predict_proba(n_estimators, clfs, approximators, X,
                            total_n_estimators, subspaces,
                            approx_flags, verbose):
    X = check_array(X)

    pred = []
    for i, subspace in enumerate(subspaces):
        estimator = clfs[i]
        if verbose > 1:
            print("predicting with estimator %d of %d for this parallel run "
                  "(total %d)..." % (i + 1, n_estimators, total_n_estimators))

        # project matrix
        X_scaled = sub_transform(X, subspace)

        # turn approximator scores to labels by outlier
        if approx_flags[i] == 1:
            raw_scores = approximators[i].predict(X_scaled)
            predicted_scores = raw_score_to_proba(estimator.decision_scores_,
                                                  raw_scores)

        else:
            predicted_scores = estimator.predict_proba(X_scaled)

        pred.append(predicted_scores[:, 1])
        # pred.append(predicted_scores)

    return pred


def _parallel_approx_estimators(n_estimators, clfs, X, total_n_estimators,
                                approx_flags, approximator, subspaces,
                                verbose):
    """
    Parameters
    ----------
    n_estimators
    clfs
    X
    total_n_estimators
    approx_flags
    approximator
    verbose

    Returns
    -------

    """
    X = check_array(X)
    # Build estimators
    approximators = []

    # TODO: approximators can be different
    for i, subspace in enumerate(subspaces):
        # project matrix
        X_scaled = sub_transform(X, subspace)

        estimator = clfs[i]

        check_is_fitted(estimator, ['decision_scores_'])
        if verbose > 1:
            print("Building estimator %d of %d for this parallel run "
                  "(total %d)..." % (i + 1, n_estimators, total_n_estimators))

        if approx_flags[i] == 1:
            # operate on the reduce space
            pseudo_scores = estimator.decision_scores_
            # pseudo_scores = estimator.decision_function(X)
            # use the same type of approximator for all models
            base_approximater = clone(approximator)
            base_approximater.fit(X_scaled, pseudo_scores)

            approximators.append(base_approximater)

        else:
            approximators.append(None)

    return approximators


idx_clf_mapping = {
    1: 'ABOD',
    2: 'CBLOF',
    3: 'FeatureBagging',
    4: 'HBOS',
    5: 'IForest',
    6: 'KNN',
    7: 'LOF',
    8: 'MCD',
    9: 'OCSVM',
    10: 'PCA',
    11: 'UNK'
}
clf_idx_mapping = {
    'ABOD': 1,
    'CBLOF': 2,
    'FeatureBagging': 3,
    'HBOS': 4,
    'IForest': 5,
    'KNN': 6,
    'LOF': 7,
    'MCD': 8,
    'OCSVM': 9,
    'PCA': 10,
    'UNK': 11
}



# -*- coding: utf-8 -*-
import os
import sys

import unittest

# temporary solution for relative imports in case pyod is not installed
# if suod is installed, no need to use the following line
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from sel_suod.models.cost_predictor import build_cost_predictor


class TestCostPredictor(unittest.TestCase):
    def setUp(self):
        pass

    def test_build(self):
        this_directory = os.path.abspath(os.path.dirname(__file__))
        build_cost_predictor(
            file_name=os.path.join(this_directory, 'saved_models',
                                   'summary_train.txt'),
            output_file="saved_models/bps_train.joblib",
            save_to_local=False)
        build_cost_predictor(
            file_name=os.path.join(this_directory, 'saved_models',
                                   'summary_prediction.txt'),
            output_file="saved_models/bps_prediction.joblib",
            save_to_local=False)

    def tearDown(self):
        pass



# -*- coding: utf-8 -*-
"""Cost predictor function for forecasting base model training and prediction
cost.
"""
# Author: Yue Zhao <zhaoy@cmu.edu>
# License: MIT
import sys
import os

import numpy as np
import pandas as pd
from joblib import dump

from scipy.stats import pearsonr
from scipy.stats import spearmanr
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error
from sklearn.metrics import r2_score
from sklearn.ensemble import RandomForestRegressor

import warnings

if not sys.warnoptions:
    warnings.simplefilter("ignore")


def indices_to_one_hot(data, nb_classes):
    """Convert an iterable of indices to one-hot encoded labels.

    Parameters
    ----------
    data : list
        The raw data.

    nb_classes : int
        The number of targeted classes.

    Returns
    -------

    """
    targets = np.array(data).reshape(-1)
    return np.eye(nb_classes)[targets]


def build_cost_predictor(file_name, output_file, save_to_local=True):
    """Build cost predictor from the scratch. In general, this does not need
    to be used.

    Parameters
    ----------
    file_name : string
        The training table of algorithm performance.

    output_file
    save_to_local
    """

    # read in the data file，
    # WS = pd.read_excel(os.path.join('saved_models', file_name),
    #                    sheet_name='sheet1').drop(columns=['File'])
    #
    # WS_np = WS.to_numpy().astype('float')
    WS_np = np.loadtxt(file_name, delimiter=',')

    X = []
    y = []

    for i in range(WS_np.shape[0]):
        for j in range(4, 14):
            X.append([WS_np[i][0], WS_np[i][1], WS_np[i][2], j - 4])
            y.append(WS_np[i][j])

        X.append([WS_np[i][0], WS_np[i][1], WS_np[i][2], j - 3])
        y.append(np.mean(y[-10:]))

    X = np.asarray(X)

    nb_classes = 11
    data = X[:, 3].astype(int)

    # build embeddings
    w = indices_to_one_hot(data, nb_classes)
    # p = indices_to_one_hot([2,3], 11)

    X = np.concatenate((X[:, [0, 1, 2]], w), axis=1)
    # this is currently a hardcoded string
    indices = [0, 1, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]
    X = X[:, indices]

    r2 = []
    mse = []
    pearson = []
    spearman = []

    # fix for 10 fold random CV now
    for i in range(10):
        X_train, X_test, y_train, y_test = train_test_split(X[:, ], y,
                                                            test_size=0.1)
        clf = RandomForestRegressor(n_estimators=100)

        clf.fit(X_train, y_train)
        test_pred = clf.predict(X_test)
        test_pred[test_pred < 0] = 0

        r2.append([r2_score(y_test, test_pred)])
        mse.append([mean_squared_error(y_test, test_pred)])
        pearson.append(pearsonr(y_test, test_pred)[0])
        spearman.append(spearmanr(y_test, test_pred)[0])

    # print('Spearman Rank', np.mean(spearman))

    clf.fit(X, y)

    if save_to_local:
        # save to the local
        dump(clf, os.path.join("saved_models", output_file))


if __name__ == "__main__":
    # this should be only executed if the pre-trained model is missing.
    build_cost_predictor(
        file_name=os.path.join('saved_models', 'summary_train.txt'),
        output_file="bps_train_curr.joblib",
        save_to_local=False)
    build_cost_predictor(
        file_name=os.path.join('saved_models', 'summary_prediction.txt'),
        output_file="bps_prediction_curr.joblib",
        save_to_local=False)



from setuptools import find_packages, setup

# read the contents of README file
from os import path
from io import open  # for Python 2 and 3 compatibility

# get __version__ from _version.py
ver_file = path.join('sel_suod', 'version.py')
with open(ver_file) as f:
    exec(f.read())

this_directory = path.abspath(path.dirname(__file__))


# read the contents of README.rst
def readme():
    with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        return f.read()


# read the contents of requirements.txt
with open(path.join(this_directory, 'requirements.txt'),
          encoding='utf-8') as f:
    requirements = f.read().splitlines()

setup(
    name='sel_suod',
    version=__version__,
    description='Fork from SUOD v0.1.3 (by Yue Zhao)',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='Jose Cribeiro',
    author_email='jose@cribeiro.net',
    url='https://github.com/jcribeiro98/Selectable_SUOD',
    download_url='https://github.com/jcribeiro98/Selectable_SUOD/archive/refs/heads/master.zip',
    keywords=['ensemble learning', 'anomaly detection', 'outlier ensembles',
              'data mining', 'machine learning', 'python'],
    packages=find_packages(exclude=['test']),
    include_package_data=True,
    install_requires=requirements,
    setup_requires=['setuptools>=38.6.0'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
